// soundblaster routines for DJGPP
// written by Dirk Jansen
// created on 23.9.1997

#include <stdio.h>
#include <stdlib.h>
#include <go32.h>
#include <dpmi.h>
#include <pc.h>
#include <dos.h>

#include "sb.h"

// global variables

int sb_baseaddr = 0x220;
int sb_irq = 5;
int sb_dmachannel = 1;

#define SB_DMABUFFERSIZE 1024;
#define SB_SAMPLEBUFFERSIZE 512; // half of SB_DMABUFFERSIZE!

const int dmabufsize = SB_DMABUFFERSIZE;
const int samplesize = SB_SAMPLEBUFFERSIZE;
const int samplefreq = 22222;
char last_error_msg[80];
int playingpart = 0;
unsigned long dmabufaddr = 0;

// local variables

static int dmabufsel;
static char nullbuffer[16384];
static int sb_running = 0;

// local function prototypes

static void autodetect ( void );

static int dspwrite ( unsigned char ch );
static unsigned char dspread ( void );
static int dspreset ( void );
static int allocatedmabuffer ( int bytes, int *sel, unsigned long *addr );
static void freedmabuffer ( int sel );
static void enablesbirq ( void );
static void disablesbirq ( void );
static void sb_irq_handler ( void );
static void installirq ( void );
static void removeirq ( void );
static void setupdma ( void );
static void setupsb ( void );

// ****************
// global functions
// ****************

// -----------------
// init soundblaster

int init_sb ( void ) {

    autodetect();
    if ( sb_irq < 0 || sb_irq > 7 ) {
        sprintf ( last_error_msg, "IRQ %d not supported!", sb_irq );
        return 0;
    }/*if*/

    if ( sb_dmachannel < 0 || sb_dmachannel > 3 ) {
        sprintf ( last_error_msg, "DMA %d not supported!", sb_dmachannel );
        return 0;
    }/*if*/

    if ( !dspreset() ) {
        sprintf ( last_error_msg, "DSP not found on port %Xh", sb_baseaddr );
        return 0;
    }/*if*/

    // setup IRQ
    installirq ();

    // allocate buffer
    if ( !allocatedmabuffer ( dmabufsize, &dmabufsel, &dmabufaddr) ) {
        strcpy ( last_error_msg, "cannot allocate dma buffer memory!" );
        removeirq ();
        dspreset ();
        return 0;
    }/*if*/

    dosmemput ( (char*) nullbuffer, dmabufsize, dmabufaddr );

    // setup dma controller
    setupdma ();

    // setup soundblaster and start transfer
    setupsb ();

    // turn on speaker
    set_sb_speaker ( 1 );

    sb_running = 1;

    return 1;

}/*init_sb*/

// -----------------
// exit soundblaster

void exit_sb ( void ) {

    if ( !sb_running ) return;

    // turn off speaker
    set_sb_speaker ( 0 );

    // reset dsp
    dspreset ();

    // remove irq
    removeirq ();

    // free dma buffer
    freedmabuffer ( dmabufsel );

    sb_running = 0;

}/*exit_sb*/

// -------------------------
// switch speaker on and off

void set_sb_speaker ( int flag ) {

    if ( flag ) {
        // enable speaker
        dspwrite ( 0xd1 );
    }/*if*/
    else {
        dspwrite ( 0xd3 );
    }/*else*/

}/*set_sb_speaker*/

// -------------------------------------
// lock memory used in interrupt handler

int lock_sb_memory ( void *ptr, int size, int code ) {

    __dpmi_meminfo meminfo;
    int seg = (code ? _my_cs() : _my_ds());

    if (__dpmi_get_segment_base_address(seg, &meminfo.address) != 0){
       return -1;
    }

    meminfo.size = size;
    meminfo.address += (long)ptr;
    return __dpmi_lock_linear_region(&meminfo);

}/*lock_memory*/

// ***************
// local functions
// ***************

// ----------------------------------
// autodetect soundblaster parameters

static void autodetect ( void ) {

    char *blaster = getenv ( "BLASTER" );

    if ( blaster ) {

        while ( *blaster ) {

             while ((*blaster == ' ') || (*blaster == '\t')) blaster++;
             if ( *blaster == 'a' || *blaster == 'A' ) {
                 sb_baseaddr = strtol(blaster+1, NULL, 16 );
             }/*if*/

             if ( *blaster == 'i' || *blaster == 'I' ) {
                 sb_irq = strtol ( blaster+1, NULL, 10 );
             }/*if*/

             if ( *blaster == 'd' || *blaster == 'D' ) {
                 sb_dmachannel = strtol ( blaster+1, NULL, 10 );
             }/*if*/

             while ((*blaster) && (*blaster != ' ') && (*blaster != '\t')) blaster++;

        }/*while*/

    }/*if*/

}/*autodetect*/

// ------------------
// write to soundcard

static int dspwrite ( unsigned char ch ) {

    int i;

    for ( i=0; i<0xffff; i++) {
        if ( !(inportb(0x0c+sb_baseaddr)&0x80)) {
            outportb(0x0c+sb_baseaddr, ch );
            return 1;
        }/*if*/
    }/*for*/

    return 0;

}/*dspwrite*/

// ------------------------------------
// read data from dsp if data available

static unsigned char dspread ( void ) {

    int i;

    for ( i=0; i<0xffff; i++) {
        if ( inportb(0x0e + sb_baseaddr) & 0x80 ) {
            return inportb(0x0a+sb_baseaddr);
        }/*if*/
    }/*for*/
    return -1;

}/*dspread*/

// -------------------
// reset soundcard dsp

static int dspreset ( void ) {

    unsigned char ch;
    int i;

    // prepare reset
    outportb ( sb_baseaddr+0x06, 1 );
    // wait
    for ( i=0; i<8; i++ ) inportb ( sb_baseaddr+0x06 );
    // do reset
    outportb ( sb_baseaddr+0x06, 0 );
    // read data (should be 0xaa)
    if ( dspread() != 0xaa ) return 0;
    return 1;

}/*dspreset*/

// ----------------------------------
// allocate dos memory for dma buffer

static int allocatedmabuffer ( int bytes, int *sel, unsigned long *addr ) {

    int seg;

    // allocate twice as much memory as we need
    seg = __dpmi_allocate_dos_memory((bytes*2+15)>>4, sel );
    if ( seg < 0 ) {
        // error!
        *sel = 0;
        *addr = 0;
        return 0;
    }/*if*/

    *addr = seg << 4;

    // look which part of memory block is on a page boundary
    if ((*addr>>16) != ((*addr+bytes)>>16)) *addr += bytes;

    return 1;

}/*allocatedmabuffer*/

// ---------------
// free dos memory

static void freedmabuffer ( int sel ) {

    __dpmi_free_dos_memory ( sel );

}/*freedmabuffer*/

// -----------------------
// enable soundblaster irq

static void enablesbirq ( void ) {

    unsigned char ch;

    ch = inportb ( 0x21 );
    ch = ch & ( ~(1 << sb_irq) );
    outportb ( 0x21, ch );

}/*enablesbirq*/

// ------------------------
// disable soundblaster irq

static void disablesbirq ( void ) {

    unsigned char ch;

    ch = inportb ( 0x21 );
    ch = ch | (1 << sb_irq );
    outportb ( 0x21, ch );

}/*disablesbirq*/

static _go32_dpmi_seginfo irq_old;   /* original prot-mode IRQ */
static _go32_dpmi_seginfo irq_new;   /* prot-mode interrupt segment info*/

// -------------------------
// irq handler for soundcard

void audio_irq ( void );

static void sb_irq_handler ( void ) {

    // acknowledge dsp
    inportb ( sb_baseaddr + 0x0e );

    // fill other part of dmabuffer with data
    playingpart = 1 - playingpart;

    audio_irq ();

    // acknowledge hardware interrupt
    outportb ( 0x20, 0x20 );

}/*sb_irq_handler*/
static int sb_irq_handler_end;

// --------------------------------
// install soundblaster irq handler

static void installirq ( void ) {

    lock_sb_memory ( sb_irq_handler, (int)sb_irq_handler_end-(int)sb_irq_handler, 1 );
    lock_sb_memory ( &playingpart, sizeof(playingpart), 0 );
    lock_sb_memory ( &sb_baseaddr, sizeof(sb_baseaddr), 0 );

    disable();
    irq_new.pm_offset = (int) sb_irq_handler;
    _go32_dpmi_allocate_iret_wrapper(&irq_new);
    irq_new.pm_selector = _go32_my_cs();
    _go32_dpmi_get_protected_mode_interrupt_vector ( sb_irq + 8, &irq_old );
    if ( _go32_dpmi_set_protected_mode_interrupt_vector ( sb_irq + 8, &irq_new ) == -1 )
        puts ( "error installing irq!" );
    enablesbirq ();
    enable();

}/*installirq*/

// -------------------------------
// remove soundblaster irq handler

static void removeirq ( void ) {

    disable();
    disablesbirq ();
    _go32_dpmi_set_protected_mode_interrupt_vector ( sb_irq + 8, &irq_old );
    _go32_dpmi_free_iret_wrapper ( &irq_new );
    enable();

}/*removeirq*/

// -----------------
// init dma transfer

static void setupdma ( void ) {

    unsigned long page = dmabufaddr >> 16;
    unsigned long offset = dmabufaddr & 0xffff;
    unsigned long samlo = (dmabufsize-1) & 0xff;
    unsigned long samhi = ( (dmabufsize-1) >> 8 ) & 0xff;
    int dma_page[8] = {0x87, 0x83, 0x81, 0x82, 0x88, 0x8b, 0x89, 0x8a };

    // master clear
    outportb ( 0x0d, 0 );
    // mask dma channel
    outportb ( 0x0a, sb_dmachannel+4 );
    // write transfer mode
    outportb ( 0x0b, 0x58 + sb_dmachannel );
    // clear flipflop
    outportb ( 0x0c, 0 );
    // write base address
    outportb ( sb_dmachannel*2, offset & 0xff );
    outportb ( sb_dmachannel*2, offset >> 8 );
    // write page number
    outportb ( dma_page[sb_dmachannel], page );
    // clear flipflop
    outportb ( 0x0c, 0 );
    // write sample length-1
    outportb ( sb_dmachannel*2+1, samlo );
    outportb ( sb_dmachannel*2+1, samhi );
    // demask channel
    outportb ( 0x0a, sb_dmachannel );

}/*setupdma*/

// -------------
// start playing

static void setupsb ( void ) {

    // generate irq every half dma buffer
    unsigned long samlo = (dmabufsize/2-1) & 0xff;
    unsigned long samhi = ( (dmabufsize/2-1) >> 8 ) & 0xff;

    // set sample rate
    dspwrite ( 0x40 );
    dspwrite ( (unsigned char)(256-1000000/samplefreq) );
    // use highspeed autoinit mode (0x90)
    dspwrite ( 0x48 );
    dspwrite ( samlo );
    dspwrite ( samhi );
    // set playmode
    dspwrite ( 0x1c );

}/*setupsb*/



